-- EuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEu
-- Program:  Chess (version 1.6a) 
-- Language: Euphoria
-- Author:   Normand M. Blais
-- Date:     August 1998
-- File:     search.e
-- EuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEuEu

include board.e
include def.e
include hash.e
include meswin.e
include movegen.e
include capture.e

sequence sm
sm = {}

integer mate
mate = 0

sequence killer
killer = repeat(0,10)

global function probettable(sequence pos,integer side,integer flag) 
  -- The format of the transposition table is
  -- ttable = {{side},{-side}}
  -- {side} = {HCODE1,HCODE2,...,HCODEk}
  -- The format of the age table is 
  -- atable = {{side},{side}}
  -- {side} = {1,2,..,n}  where 1 <= n >= 3  
  integer k
  if side = -1 then side = 2 end if                               
  k = find(pos[HCODE],ttable[side])
  if k then
     if flag then
        atable[side][k] = atable[side][k] + 1
     else
        atable[side][k] = atable[side][k] - 1
     end if
  else
     k = 1
     ttable[side] = append(ttable[side],pos[HCODE])
     atable[side] = append(atable[side],1)
  end if
  return atable[side][k]
end function

global procedure storehash(sequence pos,integer side,integer move,
                           integer score,integer bound,integer depth) 
  sequence s
  integer hashval
  hashval = pos[HASH]
  if hashval=0  or hashval > 32768 then
     return 
  end if
  if side = -1 then side = 2 end if
  if hashtable[side][hashval][CODE] != 0 then
     -- some position already in table
     if bound > hashtable[side][hashval][BOUND] and bound = 3 then
        -- exact score, update
        hashtable[side][hashval][BESTMOVE] = move
        hashtable[side][hashval][SCORE]    = score
        hashtable[side][hashval][BOUND]    = bound
        hashtable[side][hashval][DEPTH]    = depth
        hashtable[side][hashval][CODE]     = pos[HCODE]
     elsif depth > hashtable[side][hashval][DEPTH] then
        -- not exact score but higher depth, update
        hashtable[side][hashval][BESTMOVE] = move
        hashtable[side][hashval][SCORE]    = score
        hashtable[side][hashval][BOUND]    = bound
        hashtable[side][hashval][DEPTH]    = depth
        hashtable[side][hashval][CODE]     = pos[HCODE]
     end if
  else
     -- new entry
     hashtable[side][hashval][BESTMOVE] = move
     hashtable[side][hashval][SCORE]    = score
     hashtable[side][hashval][BOUND]    = bound
     hashtable[side][hashval][DEPTH]    = depth
     hashtable[side][hashval][CODE]     = pos[HCODE]
  end if        
end procedure

global function retrievehash(sequence pos,integer side)
  sequence s
  integer hashval,best,score,depth,bound
  hashval = pos[HASH]
  if hashval=0 or hashval > 32768 then
     return {}
  end if
  if side = -1 then side = 2 end if
  if hashtable[side][hashval][CODE] = pos[HCODE] then
     best = hashtable[side][hashval][BESTMOVE]
     score = hashtable[side][hashval][SCORE]
     bound = hashtable[side][hashval][BOUND]
     depth = hashtable[side][hashval][DEPTH] 
     return {best,score,bound,depth}
  end if        
  return {}
end function

-------------Recording of the top level move-list (stack)-------------------

global procedure Push_move(sequence move_rec)
  move_stack = append(move_stack,move_rec)
end procedure

global function Pop_move()
   sequence move_rec                                
   if length(move_stack) then
      move_rec = move_stack[length(move_stack)]
      move_stack = move_stack[1..length(move_stack)-1]
      return move_rec   
   end if
   return {}
end function

----------End of top level recording moves-list-----------------------------

-----------Playing moves routines-------------------------------------------

function phase(sequence pos)
  if pos[FORCES] < 4 then
     return 3 -- ending
  elsif pos[FORCES] < 8 then
     return 2 -- middle games
  end if
  return 1 -- opening
end function

function positional(integer src,integer dest,integer p,integer ph)
   if p = WP then
      return wpawnmo[dest] - wpawnmo[src]
   elsif p = BP then
      return bpawnmo[dest] - bpawnmo[src]
   elsif p = WN or p = BN then
      return knightmo[dest] - knightmo[src]
   elsif p = WB or p = BB then
      return bishopmo[dest] - bishopmo[src]
   elsif p = WR or p = BR and (ph = 2 or ph = 3) then
      return rookmo[dest] - rookmo[src]
   elsif p = WQ or p = BQ and (ph = 2 or ph = 3) then
      return queenmo[dest] - queenmo[src]
   elsif p = WK or p = BK and ph = 3 then
      return kingmo[dest] - kingmo[src]
   end if                       
   return 0
end function

constant PV = {100,300,300,500,950,7777}
--constant PV = {100,320,330,460,900,7777}

global function material(sequence pos, integer side)
  integer ma,p
  ma = 0
  for i=1 to 64 do               
     p = pos[BOARD][i]
     if side = 1 and p > 0 then 
         ma = ma + PV[p]
     elsif side = -1 and p < 0 then
         ma = ma + PV[-p]
     end if
  end for        
  return ma
end function    

global function MoveRetracter(sequence bd,integer flag)
  object x
  integer special,side,move
  integer src,dest,psrc,pdest,p 
  if bd[PLYNO] = 0 then return bd end if
  if flag then
     x        = Pop_move() 
     special  = x[SPECIAL]
     dest     = x[DEST]
     src      = x[SRC]
     pdest    = x[DENTITY]
     psrc     = x[SENTITY]
     move = dest * 1000 + src + special 
     ply = ply - 1
     movecount = floor((ply-1)/2) + 1
  else
     x       = bd[LMOVE]
     dest     = remainder(x,100)
     src      = floor(x/1000)
     special = floor(remainder(x,1000)/100) * 100
     pdest    = bd[CAPTURE]
     psrc   = bd[BOARD][src]
     move    = dest * 1000 + src + special
  end if
  if pdest then
     bd[FORCES] = bd[FORCES] + 1
     bd[PHASE] = phase(bd)
     if bd[TURN] = 1 then
        bd[WMATER] = bd[WMATER] + PV[pdest]
     else
        bd[BMATER] = bd[BMATER] + PV[-pdest]
     end if
  end if
  bd[TURN]  = -bd[TURN]
  bd[PLYNO] = bd[PLYNO] - 1
  bd[LMOVE] = move
  side = bd[TURN]
  if side = 1 then
     if move = 63061 then
        bd[BOARD][61] = WK
        bd[BOARD][64] = WR
        bd[BOARD][63] = 0
        bd[BOARD][62] = 0
        bd[CASTLING] = xor_bits(bd[CASTLING],3)
        if flag then
           Update_screen_board(bd,62064)
           Update_screen_board(bd,63061)
        end if
     elsif move = 59061 then
        bd[BOARD][61] = WK
        bd[BOARD][57] = WR
        bd[BOARD][59] = 0
        bd[BOARD][60] = 0
        bd[CASTLING] = xor_bits(bd[CASTLING],3)
        if flag then
           Update_screen_board(bd,60057)
           Update_screen_board(bd,59061)
        end if
     elsif special = 900 then
        -- Prise en passant
        bd[BOARD][src]    = WP
        bd[BOARD][dest]   = 0
        bd[BOARD][dest-8] = BP
        bd[ENPASSANT]     = 1
        if flag then
           Update_screen_board(bd,move)
           Update_screen_board(bd,(dest-8)*1000+(dest-8))
        end if
     elsif special >= 100 and special <= 400 then
        -- Promotion
        bd[BOARD][src]  = psrc
        bd[BOARD][dest] = pdest
        if flag then
           Update_screen_board(bd,move)
           Update_screen_board(bd,dest*1000+dest)
        end if
     else
        -- normal move 
        bd[BOARD][src]  = psrc
        bd[BOARD][dest] = pdest
        if flag then
           Update_screen_board(bd,move)
           Update_screen_board(bd,dest*1000+dest)
        end if
     end if
  elsif side = -1 then
     if move = 7005 then
        bd[BOARD][5] = BK
        bd[BOARD][7] = 0
        bd[BOARD][8] = BR
        bd[BOARD][6] = 0
        bd[CASTLING] = xor_bits(bd[CASTLING],12)
        if flag then
           Update_screen_board(bd,7005)
           Update_screen_board(bd,6008)
        end if
     elsif move = 3005 then
        bd[BOARD][5] = BK
        bd[BOARD][3] = 0
        bd[BOARD][1] = BR
        bd[BOARD][4] = 0
        -- august 1998
        bd[CASTLING] = xor_bits(bd[CASTLING],12)
        if flag then
--           bd[CASTLING] = xor_bits(bd[CASTLING],12)
           Update_screen_board(bd,3005)
        end if
        Update_screen_board(bd,4001)
     elsif special = 900 then
        -- Prise en passant
        bd[BOARD][src]    = BP
        bd[BOARD][dest]   = 0
        bd[BOARD][dest+8] = WP
        bd[ENPASSANT]     = 1
        if flag then
           Update_screen_board(bd,move)
           Update_screen_board(bd,(dest+8)*1000+(dest+8))
        end if
     elsif special >= 500 and special <= 800 then
        -- Promotion
        bd[BOARD][src]  = psrc
        bd[BOARD][dest] = pdest
        if flag then
           Update_screen_board(bd,move)
           Update_screen_board(bd,dest*1000+dest)
        end if
     else
        -- normal move 
        bd[BOARD][src]  = psrc
        bd[BOARD][dest] = pdest
        if flag then
           Update_screen_board(bd,move)
           Update_screen_board(bd,dest*1000+dest)
        end if
     end if 
  end if          
  if flag then
     lastmove = Pop_move()
     Push_move(lastmove)                      
     if length(moves_list) > 5 then
        moves_list = moves_list[1..length(moves_list)-4]
     end if
  end if
  bd[HASH] = remainder(hash(bd),32768)
  bd[DSCORE] = bd[DSCORE] - positional(src,dest,psrc,bd[PHASE])
  bd[HCODE] = 0
  for i=1 to length(bd[WCONTROL]) do
     bd[HCODE] = bd[HCODE] + bd[WCONTROL][i]
  end for
  for i=1 to length(bd[BCONTROL]) do
     bd[HCODE] = bd[HCODE] + bd[BCONTROL][i]
  end for
  return bd
end function

global function MoveExecutor(sequence bd,integer move,integer side,
                             integer updatescreen)
  object x
  integer special,k
  integer src,dest,psrc,pdest 
  
  bd[LMOVE] = move
  
  special = remainder(move,1000) - remainder(move,100)
  x       = move-special
  src     = floor(x/1000)
  dest    = remainder(x,100)
  psrc    = bd[BOARD][src]
  pdest   = bd[BOARD][dest]
 
  bd[CAPTURE] = pdest
  -- ??
  if not pdest  then
     bd[DSCORE] = bd[DSCORE] + positional(src,dest,psrc,bd[PHASE])
--     if pdest = WK or pdest = BK and bd[PHASE] < 3 then
--        bd[DSCORE] = bd[SCORE] - 50
--     end if
  end if
  
  if pdest then
     bd[FORCES] = bd[FORCES] - 1
     bd[PHASE] = phase(bd)
  end if                   
  
  -- In case of actual move
  if updatescreen then
     -- update move stack
     lastmove = {src,dest,special,psrc,pdest}
     Push_move(lastmove)
     -- increment the ply number
     ply = ply + 1              
     -- compute the move counter ???
     movecount = floor((ply-1)/2) + 1
     -- initiate the message buffer
     if side = 1 then
        mess = sprintf("%3d.",movecount)
     else
        mess = sprintf("%3d...",movecount)
     end if
     -- assume no pawn move, no capture
     cnt50 = cnt50 + 1                 
     -- reset if pawn move or capture
     if pdest or psrc = WP or psrc = BP then
        cnt50 = 0
     end if
  end if
  
  -- Play the move on the board (virtual or not)
  if side = 1 then
     -- check for castling moves
     if x = 61063 or x = 61059 and bd[BOARD][61] = WK then
        if x = 61063 then
           bd[BOARD][63] = bd[BOARD][61]
           bd[BOARD][61] = 0
           bd[BOARD][62] = bd[BOARD][64]
           bd[BOARD][64] = 0
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCK)
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCQ)
           bd[DSCORE] = bd[SCORE] + 50
           if updatescreen then
              Update_screen_board(bd,61063)
              Update_screen_board(bd,64062)
              moves_list = moves_list & "e1g1"
              mess = mess & "O-O"
           end if
        elsif x = 61059 then
           bd[BOARD][59] = bd[BOARD][61]
           bd[BOARD][61] = 0
           bd[BOARD][60] = bd[BOARD][57]
           bd[BOARD][57] = 0
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCK)
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCQ)
           bd[DSCORE] = bd[SCORE] + 50
           if updatescreen then
              Update_screen_board(bd,61059)
              Update_screen_board(bd,57060)
              moves_list = moves_list & "e1c1"
              mess = mess & "O-O-O"
           end if
        end if
     elsif special = 900 then
        -- Prise en passant
        bd[BOARD][dest] = WP
        bd[BOARD][src] = 0
        bd[BOARD][dest+8] = 0
        bd[ENPASSANT] = 0
        if updatescreen then
           Update_screen_board(bd,x)
           Update_screen_board(bd,(dest+8)*1000+(dest+8))
           moves_list = moves_list & Move2text(x)
           mess = mess & Move2Text(bd,x)
        end if
     elsif special >= 100 and special <= 400 then
        -- Promotion
        bd[BOARD][dest] = PROMOTION[floor(special/100)]
        bd[BOARD][src] = 0
        if updatescreen then
           Update_screen_board(bd,x)
           moves_list = moves_list & Move2text(x)
           mess = mess & Move2Text(bd,x)
        end if
     else
        -- normal move 
        bd[BOARD][dest] = psrc
        bd[BOARD][src] = 0
        if updatescreen then
           Update_screen_board(bd,x)
           moves_list = moves_list & Move2text(x)
           mess = mess & Move2Text(bd,x)
        end if
     end if
  elsif side = -1 then
     if x = 5007 or x = 5003  and bd[BOARD][5] = BK then
        if x = 5007 then
           bd[BOARD][7] = bd[BOARD][5]
           bd[BOARD][5] = 0
           bd[BOARD][6] = bd[BOARD][8]
           bd[BOARD][8] = 0
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCK)
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCQ)
           bd[DSCORE] = bd[SCORE] + 50
           if updatescreen then
              Update_screen_board(bd,5007)
              Update_screen_board(bd,8006)
              moves_list = moves_list & "e8g8"
              mess = mess & "O-O"
           end if
        elsif x = 5003 then
           bd[BOARD][3] = bd[BOARD][5]
           bd[BOARD][5] = 0
           bd[BOARD][4] = bd[BOARD][1]
           bd[BOARD][1] = 0
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCK)
           bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCQ)
           bd[DSCORE] = bd[SCORE] + 50
           if updatescreen then
              Update_screen_board(bd,5003)
              Update_screen_board(bd,1004)
              moves_list = moves_list & "e8c8"
              mess = mess & "O-O-O"
           end if
        end if
     elsif special = 900 then
        -- Prise en passant
        bd[BOARD][dest] = BP
        bd[BOARD][src] = 0
        bd[BOARD][dest-8] = 0
        bd[ENPASSANT] = 0
        if updatescreen then
           Update_screen_board(bd,x)
           Update_screen_board(bd,(dest-8)*1000+(dest-8))
           moves_list = moves_list & Move2text(x)
           mess = mess & Move2Text(bd,x)
        end if
     elsif special >= 500 and special <= 800 then
        -- Promotion
        bd[BOARD][dest] = PROMOTION[floor(special/100)]
        bd[BOARD][src] = 0
        if updatescreen then
           Update_screen_board(bd,x)
           moves_list = moves_list & Move2text(x)
           mess = mess & Move2Text(bd,x)
        end if
     else
        -- normal move 
        bd[BOARD][dest] = psrc
        bd[BOARD][src] = 0
        if updatescreen then
           Update_screen_board(bd,x)
           moves_list = moves_list & Move2text(x)
           mess =  mess & Move2Text(bd,x)
        end if
     end if
  end if
-----------------------------------------------------------------------
-- Update board record status
-----------------------------------------------------------------------
-- En passant
-----------------------------------------------------------------------
  bd[ENPASSANT] = 0
  if special = 0 and psrc = WP then
     bd[ENPASSANT] = Pawnhasmovedtwo(bd,src,dest)
  elsif special = 0 and psrc = BP then
     bd[ENPASSANT] = Pawnhasmovedtwo(bd,src,dest)
  end if
  if bd[CAPTURE] then
     if side = 1 then
        bd[BMATER] = bd[BMATER] - PV[-pdest]
     else
        bd[WMATER] = bd[WMATER] - PV[pdest]
     end if
  end if
-----------------------------------------------------------------------
-- King has moved
-----------------------------------------------------------------------
  -- "side" has just moved, update position status
  if psrc = WK then
     if and_bits(bd[CASTLING],MASK_WCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCK)
     end if
     if and_bits(bd[CASTLING],MASK_WCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCQ)
     end if  
  elsif psrc = BK then
     if and_bits(bd[CASTLING],MASK_BCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCK)
     end if
     if and_bits(bd[CASTLING],MASK_BCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCQ)
     end if
  end if
-----------------------------------------------------------------------
-- Opponent's Rook has been captured
-----------------------------------------------------------------------
  -- "side" has just moved, update position status
  if bd[CAPTURE] = WR then
     if dest = 64 and and_bits(bd[CASTLING],MASK_WCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCK)
     elsif dest = 57 and and_bits(bd[CASTLING],MASK_WCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCQ)
     end if  
  elsif bd[CAPTURE] = BR then
     if dest = 8 and and_bits(bd[CASTLING],MASK_BCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCK)
     elsif dest = 1 and and_bits(bd[CASTLING],MASK_BCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCQ)
     end if
  end if
  if psrc = WR then
     -- assume king's rook
     if bd[BOARD][57] = WR and and_bits(bd[CASTLING],MASK_WCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCK)
     -- else assume queen's rook 
     elsif bd[BOARD][64] = WR and and_bits(bd[CASTLING],MASK_WCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCQ)
     -- else it does not matter
     elsif and_bits(bd[CASTLING],MASK_WCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCK)
     elsif and_bits(bd[CASTLING],MASK_WCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_WCQ)
     end if  
  elsif psrc = BR then
     -- assume king's rook
     if bd[BOARD][1] = BR and and_bits(bd[CASTLING],MASK_BCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCK)
     -- else assume queen's rook 
     elsif bd[BOARD][8] = BR and and_bits(bd[CASTLING],MASK_BCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCQ)
     -- else it does not matter
     elsif and_bits(bd[CASTLING],MASK_BCK) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCK)
     elsif and_bits(bd[CASTLING],MASK_BCQ) then
        bd[CASTLING] = xor_bits(bd[CASTLING],MASK_BCQ)
     end if  
  end if
-----------------------------------------------------------------------
-- Control
-----------------------------------------------------------------------
  if side = 1 then  
     bd[WCONTROL] = Controlgen(bd,1)
     x = find(BK,bd[BOARD])
     bd[CHECK] = (find(x,bd[WCONTROL]) > 0)
  elsif side = -1 then
     bd[BCONTROL] = Controlgen(bd,-1)
     x = find(WK,bd[BOARD])        
     bd[CHECK] = (find(x,bd[BCONTROL]) > 0)
  end if 
-----------------------------------------------------------------------
-----------------------------------------------------------------------
  bd[TURN] = -bd[TURN]
  bd[PLYNO] = bd[PLYNO] + 1
  bd[HASH] = remainder(hash(bd),32768)
  bd[HCODE] = 0
  for i=1 to length(bd[WCONTROL]) do
     bd[HCODE] = bd[HCODE] + bd[WCONTROL][i]
  end for
  for i=1 to length(bd[BCONTROL]) do
     bd[HCODE] = bd[HCODE] + bd[BCONTROL][i]
  end for
-----------------------------------------------------------------------
-----------------------------------------------------------------------
  if updatescreen then
     if bd[CHECK] then
        mess = mess & "+"
     end if
     Writemessage(mess)
  end if
  return bd
end function


-------------------End of playing moves routine-----------------------------

---------------Scoring routines---------------------------------------------

global function IsMate(sequence pos,integer side)
  object x
  x = MovesGenerator(pos,side)
  mate = 0
--  if length(x[1]) = 0 and length(x[2]) = 0 then
  if length(x) = 0 then
--     if pos[CHECK] then          
        mate = 1
        return 1
--     end if
  end if          
  return 0
end function

function i2c(integer m)
  integer c
  c = remainder(m,8)
  if c = 0 then
     c = 8
  end if
--  r = floor((m-1)/8) + 1
  return c
end function 

constant colsort = { 
{ 1, 9,17,25,33,41,49,57},
{ 2,10,18,26,34,42,50,58},
{ 3,11,19,27,35,43,51,59},
{ 4,12,20,28,36,44,52,60},
{ 5,13,21,29,37,45,53,61},
{ 6,14,22,30,38,46,54,62},
{ 7,15,23,31,39,47,55,63},
{ 8,16,24,32,40,48,56,64}
}

function doublepawn(sequence pos, integer side)
  integer c,db,p
  if side = 1 then p = WP else p = BP end if
  -- double pawn
  db = 0
  for i=1 to 8 do
     c = 0 
     for j=1 to 8 do
        if pos[BOARD][colsort[i][j]] = p then
           c = c + 1
        end if
     end for   
     if c > 0 then       
        db = db + c - 1
     end if
  end for
  return db
end function

-- function isolatedpawn(sequence pos, integer side)
--   integer c,p
--   if side = 1 then p = WP else p = BP end if
--   c = 0
--   for i=1 to 8 do
--   for j=1 to 8 do
--      if pos[BOARD][colsort[i][j]] = p then
--         if i = 1 then
--            if find(p,colsort[i+1]) = 0 then
--               c = c + 2
--            end if
--         elsif i = 8 then
--            if find(p,colsort[i-1]) = 0 then
--               c = c + 2
--            end if      
--         else
--            if find(p,colsort[i+1]) = 0 and 
--               find(p,colsort[i-1]) = 0 then
--               c = c + 1
--            end if      
--         end if
--      end if
--   end for
--   end for
--   return c
-- end function

-- function advancedpawn(sequence pos, integer side)
--   integer c,p
--   if side = 1 then p = WP else p = BP end if
--   c = 0
--   for i=1 to 8 do
--   for j=1 to 8 do
--      if pos[BOARD][colsort[i][j]] = p then
--         if j < 5 and side = 1 then
--            c = c + (5-j)
--         elsif j >= 5 and side = -1 then
--            c = c + (j-4)
--         end if
--      end if
--   end for
--   end for
--   return c
-- end function
                      

-- function protected(sequence pos,integer side,integer square)
--   if find(square,pos[WCONTROL]) and side = 1 then
--   return 1     
--   elsif find(square,pos[BCONTROL]) and side = -1 then
--   return 1     
--   end if                                       
--   return 0
-- end function

-- function unprotected(sequence pos,integer side, integer square)
--   if not protected(pos,side,square) then
--   return 1
--   end if     
--   return 0
-- end function


function Material(sequence bd, integer side)
  integer ma1,ma2,p
  ma1 = 0  ma2 = 0
  for i=1 to 64 do               
     p = bd[i]
     if p > 0 then 
         ma1 = ma1 + PV[p]
     elsif p < 0 then
         ma2 = ma2 + PV[-p]
     end if
  end for        
  if side = 1 then 
     return ma1 - ma2
  end if   
  return ma2 - ma1
end function    

global function examine(sequence pos,integer side,sequence mv)
  sequence tmp
  integer k,sq,loc
  if pos[CAPTURE] then
     sq = remainder(pos[LMOVE],100)
     tmp = remainder(mv,100)
     for i=1 to length(tmp) do
        if tmp[i] = sq then
           mv = mv[i] & mv[1..i-1] & mv[i+1..length(mv)]
        end if
     end for
  end if
  if side = 1 and pos[PHASE] < 3 then 
     tmp = floor(mv/1000)
     for i=1 to length(tmp) do
        if pos[BOARD][tmp[i]] = WK then
           mv = mv[1..i-1] & mv[i+1..length(mv)] & mv[i]
        end if
     end for
  elsif side = -1 and pos[PHASE] < 3 then
     tmp = floor(mv/1000)
     for i=1 to length(tmp) do
        if pos[BOARD][tmp[i]] = BK then
           mv = mv[1..i-1] & mv[i+1..length(mv)] & mv[i]
        end if
     end for
  end if                      
  if side = 1 and pos[PHASE] < 3 and pos[BOARD][61] = WK then 
     k = find(61059,mv)
     if k then
        mv = mv[k] & mv[1..k-1] & mv[k+1..length(mv)]
     end if
     k = find(61063,mv)
     if k then
        mv = mv[k] & mv[1..k-1] & mv[k+1..length(mv)]
     end if
  elsif side = -1 and pos[PHASE] < 3 and pos[BOARD][5] = BK then
     k = find(5003,mv)
     if k then
        mv = mv[k] & mv[1..k-1] & mv[k+1..length(mv)]
     end if
     k = find(5007,mv)
     if k then
        mv = mv[k] & mv[1..k-1] & mv[k+1..length(mv)]
     end if
  end if                      
  tmp = remainder(mv,1000)
  for i= 1 to length(tmp) do
     if tmp[i] > 100 and tmp[i] < 900 then
        mv = mv[i] & mv[1..i-1] & mv[i+1..length(mv)]
     end if
  end for
  return mv
end function

constant KK = {-9,-8,-7,-1,+1,+7,+8,+9}

function kingsafety(sequence pos,integer side)
  -- check if squares surounding the king are under attack
  integer loc,sum,p
  sequence s,c
  if side = 1 then p = WK c = pos[BCONTROL] 
  else p = BK c = pos[WCONTROL] end if
  loc = find(p,pos[BOARD])
  s = {}
  for i=1 to length(KK) do
     if loc+i >= 1 and loc+i <= 64 then
        s = append(s,loc+i)
     end if
  end for
  sum = 0
  for k=1 to length(s) do 
     for i=1 to length(c) do
        if s[k] = c[i] then
           sum = sum + 1     
        end if
     end for                 
  end for
  return sum 
end function

----------------------------------------------------------------------------

global function Score(sequence pos,integer side)
  integer ma,cntrl,safety,structure,tactic,v
  integer xside
  xside = -side
  v = 0
  if side = 1 then
     cntrl = floor((length(pos[WCONTROL]) - length(pos[BCONTROL]))/3)
     ma = pos[WMATER] - pos[BMATER]
  else
     cntrl = floor((length(pos[BCONTROL]) - length(pos[WCONTROL]))/3)
     ma = pos[BMATER] - pos[WMATER]
  end if
  safety = kingsafety(pos,side) - kingsafety(pos,xside)
  structure = doublepawn(pos, side) - doublepawn(pos, xside)
  tactic = 0
  v = ma + cntrl + safety + structure + tactic
  return v
end function

---------End of scoring routines--------------------------------------------

---------Part of moves generation routines----------------------------------


global function Successors(sequence pos,integer side,integer flag,
                           integer ply)
  object x 
  sequence moves,si,set
  integer k
  x = {}
  if flag then
     x = MovesGenerator(pos,side)               
     if length(x) = 0 then mate=1 return {} end if
--     moves = x[1] & x[2]
     moves = x
  else
--     x = XMoveGenerator(pos,side)
     x = CaptureGenerator(pos,side)
     if length(x) = 0 then mate=1 return {} end if
--     moves = x[1] & x[2]
     moves = x
  end if
  moves = examine(pos,side,moves)
--  k = find(killer[ply],moves)
  k = find(killer[ply],moves)
  if k > 1 then
     moves = killer[ply] & moves[1..k-1] & moves[k+1..length(moves)]
  end if
  set = {}     
  for i=1 to length(moves) do
     set = append(set,MoveExecutor(pos,moves[i],side,0))
  end for
  return set 
end function

--------End of moves generation routines------------------------------------

-------utility functions----------------------------------------------------

function Empty(object x)
  if sequence(x) then
    if length(x) then
      return 0 -- False
    end if
  end if
  return 1 -- True
end function     

global procedure showmove(sequence pos,integer d, integer line)
  integer p
  p = pos[BOARD][remainder(pos[LMOVE],100)]
  sm = Move2text(pos[LMOVE])
  if pos[CAPTURE] and  p > 1 then
     sm = PIECES[p] & sm[1..2] & 'x' & sm[3..4]
  elsif pos[CAPTURE] and  p < -1 then
     sm = PIECES[-p] & sm[1..2] & 'x' & sm[3..4]
  elsif pos[CAPTURE] and  p = 1 then
     sm = PIECES[p] & sm[1..2] & 'x' & sm[3..4]
  elsif pos[CAPTURE] and  p = -1 then
     sm = PIECES[-p] & sm[1..2] & 'x' & sm[3..4]
  elsif p > 1 then
     sm = PIECES[p] & sm[1..2] & '-' & sm[3..4]
  elsif p < -1 then
     sm = PIECES[-p] & sm[1..2] & '-' & sm[3..4]
  else
     sm = " " & sm[1..2] & '-' & sm[3..4]
  end if         
  if d = 0 then d = 1 end if
  position(line,d*7)
  printf(1,"%s", {sm & "                                              "})
end procedure

global function cnvtmove(sequence pos)
  integer p
  p = pos[BOARD][remainder(pos[LMOVE],100)]
  sm = Move2text(pos[LMOVE])
  if pos[CAPTURE] and  p > 1 then
     sm = PIECES[p] & sm[1..2] & 'x' & sm[3..4]
  elsif pos[CAPTURE] and  p < -1 then
     sm = PIECES[-p] & sm[1..2] & 'x' & sm[3..4]
  elsif pos[CAPTURE] and  p = 1 then
     sm = PIECES[p] & sm[1..2] & 'x' & sm[3..4]
  elsif pos[CAPTURE] and  p = -1 then
     sm = PIECES[-p] & sm[1..2] & 'x' & sm[3..4]
  elsif p > 1 then
     sm = PIECES[p] & sm[1..2] & '-' & sm[3..4]
  elsif p < -1 then
     sm = PIECES[-p] & sm[1..2] & '-' & sm[3..4]
  else
     sm = " " & sm[1..2] & '-' & sm[3..4]
  end if    
  return sm
end function

---------------end of utility functions-------------------------------------

-----------Search functions-------------------------------------------------

global function srch(sequence pos, integer side, integer ply,
                       integer d, integer alpha, integer beta)
  object x                     
  sequence succ
  integer val,a,b,len,best
  integer xside,flag
  flag = 0
  xside = -side
  if get_key() = 27 then
     if graphics_mode(-1) then
     end if
     abort(0)
  end if
  if pos[CHECK] then
     if IsMate(pos,xside) then
        return 200000
     end if
  end if
  mainline[try][ply] = cnvtmove(pos)
  if pos[CAPTURE] then
     if d < dmax then
        d = d + 2
        flag = 1
     end if
  end if
  if ply >= d then
     val = Score(pos,side)
     storehash(pos,side,pos[LMOVE],val,3,ply) 
     return val   
  end if
  a = alpha
  b = beta
--  best = -INFINITY
  best = a
  succ = Successors(pos,xside,flag,ply)
  len = length(succ)
  for i=1 to len do
     if best > b then 
        killer[ply] = succ[i][LMOVE] 
        storehash(succ[i],xside,succ[i][LMOVE],b,remainder(ply,2)+1,ply) 
        exit 
     end if  
     if best > a then a = best end if
--     val = -search(succ[i],xside,ply+1,d,-b,-a)
     if best < val then best = val end if
  end for
  return best
end function


global function search(sequence pos, integer side, integer ply,
                       integer d, integer alpha, integer beta)
  object x                     
  sequence succ
  integer val,a,b,len
  integer xside,flag
  flag = 0
  xside = -side
  if get_key() = 27 then
     if graphics_mode(-1) then
     end if
     abort(0)
  end if
  if pos[CHECK] then
     if IsMate(pos,xside) then
        return 200000
     end if
  end if
  mainline[try][ply] = cnvtmove(pos)
  if pos[CAPTURE] or pos[CHECK] then
     if d < dmax then
        d = d + 2
        flag = 1
     end if
  end if
  if ply >= d then
     val = Score(pos,side) - zscore
     storehash(pos,side,pos[LMOVE],val,3,ply) 
     return val   
  end if
  a = alpha
  b = beta
  val = -INFINITY
  succ = Successors(pos,xside,flag,ply)
  len = length(succ)
  if remainder(ply,2) != 0 then   -- min node
     for i=1 to len do
        x = retrievehash(succ[i],xside)
        if length(x) then
           if x[DEPTH] >= ply and x[BOUND] = 3 then
              b = x[SCORE]
              if b < 0 then b = -b end if
           end if
        else
           val = search(succ[i],xside,ply+1,d,a,b)
        end if
        if b > val then b = val end if
        if b <= a then 
           killer[ply] = succ[i][LMOVE] 
           storehash(succ[i],xside,succ[i][LMOVE],b,2,ply) 
           exit 
        end if
     end for
     return b
  else  -- max node
     for i=1 to len do
        x = retrievehash(succ[i],xside)
        if length(x) then
           if x[DEPTH] >= ply and x[BOUND] = 3 then
              a = x[SCORE]               
              if a > 0 then a = -a end if
           end if
        else
           val = search(succ[i],xside,ply+1,d,a,b)
        end if
        if a < val then a = val end if
        if a >= b then  
           killer[ply] = succ[i][LMOVE] 
           storehash(succ[i],xside,succ[i][LMOVE],a,1,ply) 
           exit 
        end if
     end for
     return a
  end if
  return {}
end function

----------End of search functions-------------------------------------------

